"""
_RunJobByStatus_

Oracle implementation of Monitoring.RunJobByStatus
"""
from __future__ import print_function, division

from WMCore.BossAir.MySQL.RunJobByStatus import RunJobByStatus as MySQLRunJobByStatus

class RunJobByStatus(MySQLRunJobByStatus):
    pass
