#!/usr/bin/env python
"""
_LoadBulkFilesByID_

Oracle implementation of DBSBufferFiles.LoadBulkFilesByID
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.LoadBulkFilesByID import LoadBulkFilesByID as MySQLLoadBulkFilesByID

class LoadBulkFilesByID(MySQLLoadBulkFilesByID):
    """
    Same as MySQL

    """
    pass
