#!/usr/bin/env python
"""
_LoadFromIDWithType_

Oracle implementation of LoadFromIDWithType

Created on Oct 3, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.LoadFromIDWithType import LoadFromIDWithType as MySQLLoadFromIDWithType

class LoadFromIDWithType(MySQLLoadFromIDWithType):
    pass
