#!/usr/bin/env python
"""
_Generators_

Generator implementations to be passed to a JobFactory instance in order
to generate custom quantities for jobs outside of the normal
WMBS Job Splitting quantities

"""
__all__ = []
