#!/usr/bin/env python
"""
_Update_

Oracle implementation of Files.Update

Created on Feb 25, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Files.Update import Update as MySQLUpdate

class Update(MySQLUpdate):
    pass
